// file      : xsd/cxx/parser/elements.txx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2005 Code Synthesis Tools CC
// license   : GNU GPL v2 + exceptions; see accompanying LICENSE file

#include <istream>

#include <xercesc/sax2/Attributes.hpp>
#include <xercesc/sax2/XMLReaderFactory.hpp>

#include <xsd/cxx/xml/string.hxx>
#include <xsd/cxx/xml/elements.hxx>
#include <xsd/cxx/xml/sax/error-handler.hxx>
#include <xsd/cxx/xml/sax/std-input-source.hxx>

namespace xsd
{
  namespace cxx
  {
    namespace parser
    {
      // sax_event_router
      //

      template <typename C>
      void sax_event_router<C>::
      startElement(XMLCh const* const uri,
                   XMLCh const* const lname,
                   XMLCh const* const /*qname*/,
                   xercesc::Attributes const& attributes)
      {
        string ns (xml::transcode<C> (uri));
        string name (xml::transcode<C> (lname));

        t_.start_element (ns, name);

        for (unsigned long i (0); i < attributes.getLength(); ++i)
        {
          string ns (xml::transcode<C> (attributes.getURI (i)));

          string name (xml::transcode<C> (attributes.getLocalName (i)));

          string value (xml::transcode<C> (attributes.getValue (i)));

          t_.attribute (ns, name, value);
        }
      }

      template <typename C>
      void sax_event_router<C>::
      endElement(XMLCh const* const uri,
                 XMLCh const* const lname,
                 XMLCh const* const /*qname*/)
      {
        string ns (xml::transcode<C> (uri));
        string name (xml::transcode<C> (lname));

        if (!buf_.empty ())
        {
          t_.characters (buf_);
          buf_.clear ();
        }

        t_.end_element (ns, name);
      }

      template <typename C>
      void sax_event_router<C>::
      characters (XMLCh const* const s, unsigned int const length)
      {
        if (length == 0)
          return;

        buf_ += xml::transcode<C> (s, length);
      }


      // traverser_impl
      //

      template <typename C>
      bool traverser_impl<C>::
      start_element_ (std::basic_string<C> const& ns,
                      std::basic_string<C> const& n,
                      traverser_base<C>* cur)
      {
        bool r (false);

        if (count_ != 0)
        {
          cur_->start_element (ns, n);
        }
        else
        {
          cur_ = cur;
          r = true;
        }

        ++count_;

        return r;
      }

      template <typename C>
      bool traverser_impl<C>::
      end_element_ (std::basic_string<C> const& ns,
                    std::basic_string<C> const& n)
      {
        bool r (false);

        --count_;

        if (count_ != 0)
        {
          cur_->end_element (ns, n);
        }
        else
        {
          cur_ = 0;
          r = true;
        }

        return r;
      }

      // traverser
      //

      // parse (uri)
      //


      template <typename X, typename C>
      X traverser<X, C>::
      parse (std::basic_string<C> const& uri,
             std::basic_string<C> const& root_element_name)
      {
        xml::auto_initializer initializer;

        std::auto_ptr<xercesc::SAX2XMLReader> sax (create_sax ());

        return parse (uri, root_element_name, *sax);
      }


      template <typename X, typename C>
      X traverser<X, C>::
      parse (C const* uri,
             std::basic_string<C> const& root_element_name)
      {
        return parse (std::basic_string<C> (uri), root_element_name);
      }


      template <typename X, typename C>
      X traverser<X, C>::
      parse (std::basic_string<C> const& uri,
             std::basic_string<C> const& root_element_name,
             xercesc::ErrorHandler& eh)
      {
        std::auto_ptr<xercesc::SAX2XMLReader> sax (create_sax ());

        sax->setErrorHandler (&eh);

        return parse (uri, root_element_name, *sax);
      }


      template <typename X, typename C>
      X traverser<X, C>::
      parse (C const* uri,
             std::basic_string<C> const& root_element_name,
             xercesc::ErrorHandler& eh)
      {
        return parse (std::basic_string<C> (uri), root_element_name, eh);
      }


      template <typename X, typename C>
      X traverser<X, C>::
      parse (std::basic_string<C> const& uri,
             std::basic_string<C> const& root_element_name,
             xercesc::SAX2XMLReader& sax)
      {
        using namespace xercesc;

        document<C> doc (*this, root_element_name);
        sax_event_router<C> router (doc);

        ErrorHandler* old_eh (sax.getErrorHandler ());
        xml::sax::error_handler<C> eh (old_eh);

        try
        {
          sax.setErrorHandler (&eh);

          sax.setContentHandler (&router);
          sax.parse (xml::string (uri).c_str ());

          sax.setErrorHandler (old_eh);
        }
        catch (...)
        {
          sax.setErrorHandler (old_eh);
          throw;
        }

        if (eh.failed ())
          throw invalid_instance ();

        return post ();
      }

      template <typename X, typename C>
      X traverser<X, C>::
      parse (C const* uri,
             std::basic_string<C> const& root_element_name,
             xercesc::SAX2XMLReader& sax)
      {
        return parse (std::basic_string<C> (uri), root_element_name, sax);
      }


      // parse (istream)
      //


      template <typename X, typename C>
      X traverser<X, C>::
      parse (std::istream& is,
             std::basic_string<C> const& root_element_name)
      {
        xml::auto_initializer initializer;

        xml::sax::std_input_source isrc (is);

        return parse (isrc, root_element_name);
      }

      template <typename X, typename C>
      X traverser<X, C>::
      parse (std::istream& is,
             std::basic_string<C> const& root_element_name,
             xercesc::ErrorHandler& eh)
      {
        xml::auto_initializer initializer;

        xml::sax::std_input_source isrc (is);

        return parse (isrc, root_element_name, eh);
      }

      template <typename X, typename C>
      X traverser<X, C>::
      parse (std::istream& is,
             std::basic_string<C> const& root_element_name,
             xercesc::SAX2XMLReader& sax)
      {
        xml::auto_initializer initializer;

        xml::sax::std_input_source isrc (is);

        return parse (isrc, root_element_name, sax);
      }


      // parse (istream, system_id)
      //


      template <typename X, typename C>
      X traverser<X, C>::
      parse (std::istream& is,
             std::basic_string<C> const& system_id,
             std::basic_string<C> const& root_element_name)
      {
        xml::auto_initializer initializer;

        xml::sax::std_input_source isrc (is, system_id);

        return parse (isrc, root_element_name);
      }


      template <typename X, typename C>
      X traverser<X, C>::
      parse (std::istream& is,
             std::basic_string<C> const& system_id,
             std::basic_string<C> const& root_element_name,
             xercesc::ErrorHandler& eh)
      {
        xml::auto_initializer initializer;

        xml::sax::std_input_source isrc (is, system_id);

        return parse (isrc, root_element_name, eh);
      }

      template <typename X, typename C>
      X traverser<X, C>::
      parse (std::istream& is,
             std::basic_string<C> const& system_id,
             std::basic_string<C> const& root_element_name,
             xercesc::SAX2XMLReader& sax)
      {
        xml::auto_initializer initializer;

        xml::sax::std_input_source isrc (is, system_id);

        return parse (isrc, root_element_name, sax);
      }

      // parse (istream, system_id, public_id)
      //

      template <typename X, typename C>
      X traverser<X, C>::
      parse (std::istream& is,
             std::basic_string<C> const& system_id,
             std::basic_string<C> const& public_id,
             std::basic_string<C> const& root_element_name)
      {
        xml::auto_initializer initializer;

        xml::sax::std_input_source isrc (is, system_id, public_id);

        return parse (isrc, root_element_name);
      }

      template <typename X, typename C>
      X traverser<X, C>::
      parse (std::istream& is,
             std::basic_string<C> const& system_id,
             std::basic_string<C> const& public_id,
             std::basic_string<C> const& root_element_name,
             xercesc::ErrorHandler& eh)
      {
        xml::auto_initializer initializer;

        xml::sax::std_input_source isrc (is, system_id, public_id);

        return parse (isrc, root_element_name, eh);
      }

      template <typename X, typename C>
      X traverser<X, C>::
      parse (std::istream& is,
             std::basic_string<C> const& system_id,
             std::basic_string<C> const& public_id,
             std::basic_string<C> const& root_element_name,
             xercesc::SAX2XMLReader& sax)
      {
        xml::auto_initializer initializer;

        xml::sax::std_input_source isrc (is, system_id, public_id);

        return parse (isrc, root_element_name, sax);
      }


      // parse (InputSource)
      //


      template <typename X, typename C>
      X traverser<X, C>::
      parse (xercesc::InputSource const& is,
             std::basic_string<C> const& root_element_name)
      {
        std::auto_ptr<xercesc::SAX2XMLReader> sax (create_sax ());

        return parse (is, root_element_name, *sax);
      }


      template <typename X, typename C>
      X traverser<X, C>::
      parse (xercesc::InputSource const& is,
             std::basic_string<C> const& root_element_name,
             xercesc::ErrorHandler& eh)
      {
        std::auto_ptr<xercesc::SAX2XMLReader> sax (create_sax ());

        sax->setErrorHandler (&eh);

        return parse (is, root_element_name, *sax);
      }


      template <typename X, typename C>
      X traverser<X, C>::
      parse (xercesc::InputSource const& is,
             std::basic_string<C> const& root_element_name,
             xercesc::SAX2XMLReader& sax)
      {
        using namespace xercesc;

        document<C> doc (*this, root_element_name);
        sax_event_router<C> router (doc);

        ErrorHandler* old_eh (sax.getErrorHandler ());
        xml::sax::error_handler<C> eh (old_eh);

        try
        {
          sax.setErrorHandler (&eh);

          sax.setContentHandler (&router);
          sax.parse (is);

          sax.setErrorHandler (old_eh);
        }
        catch (...)
        {
          sax.setErrorHandler (old_eh);
          throw;
        }

        if (eh.failed ())
          throw invalid_instance ();

        return post ();
      }


      template <typename X, typename C>
      std::auto_ptr<xercesc::SAX2XMLReader> traverser<X, C>::
      create_sax ()
      {
        using namespace xercesc;

        std::auto_ptr<SAX2XMLReader> sax (
          XMLReaderFactory::createXMLReader());

        sax->setFeature(XMLUni::fgSAX2CoreValidation, true);
        sax->setFeature(XMLUni::fgXercesDynamic, false);
        sax->setFeature(XMLUni::fgSAX2CoreNameSpaces, true);
        sax->setFeature(XMLUni::fgXercesSchema, true);
        sax->setFeature(XMLUni::fgXercesSchemaFullChecking, true);
        sax->setFeature(XMLUni::fgSAX2CoreNameSpacePrefixes, true);
        sax->setFeature (XMLUni::fgXercesValidationErrorAsFatal, true);

        return sax;
      }


      // traverser<void>
      //

      // parse (uri)
      //


      template <typename C>
      void traverser<void, C>::
      parse (std::basic_string<C> const& uri,
             std::basic_string<C> const& root_element_name)
      {
        xml::auto_initializer initializer;

        std::auto_ptr<xercesc::SAX2XMLReader> sax (create_sax ());

        parse (uri, root_element_name, *sax);
      }


      template <typename C>
      void traverser<void, C>::
      parse (C const* uri,
             std::basic_string<C> const& root_element_name)
      {
        parse (std::basic_string<C> (uri), root_element_name);
      }


      template <typename C>
      void traverser<void, C>::
      parse (std::basic_string<C> const& uri,
             std::basic_string<C> const& root_element_name,
             xercesc::ErrorHandler& eh)
      {
        std::auto_ptr<xercesc::SAX2XMLReader> sax (create_sax ());

        sax->setErrorHandler (&eh);

        parse (uri, root_element_name, *sax);
      }


      template <typename C>
      void traverser<void, C>::
      parse (C const* uri,
             std::basic_string<C> const& root_element_name,
             xercesc::ErrorHandler& eh)
      {
        parse (std::basic_string<C> (uri), root_element_name, eh);
      }


      template <typename C>
      void traverser<void, C>::
      parse (std::basic_string<C> const& uri,
             std::basic_string<C> const& root_element_name,
             xercesc::SAX2XMLReader& sax)
      {
        using namespace xercesc;

        document<C> doc (*this, root_element_name);
        sax_event_router<C> router (doc);

        ErrorHandler* old_eh (sax.getErrorHandler ());
        xml::sax::error_handler<C> eh (old_eh);

        try
        {
          sax.setErrorHandler (&eh);

          sax.setContentHandler (&router);
          sax.parse (xml::string (uri).c_str ());

          sax.setErrorHandler (old_eh);
        }
        catch (...)
        {
          sax.setErrorHandler (old_eh);
          throw;
        }

        if (eh.failed ())
          throw invalid_instance ();

        post ();
      }

      template <typename C>
      void traverser<void, C>::
      parse (C const* uri,
             std::basic_string<C> const& root_element_name,
             xercesc::SAX2XMLReader& sax)
      {
        parse (std::basic_string<C> (uri), root_element_name, sax);
      }


      // parse (istream)
      //


      template <typename C>
      void traverser<void, C>::
      parse (std::istream& is,
             std::basic_string<C> const& root_element_name)
      {
        xml::auto_initializer initializer;

        xml::sax::std_input_source isrc (is);

        parse (isrc, root_element_name);
      }

      template <typename C>
      void traverser<void, C>::
      parse (std::istream& is,
             std::basic_string<C> const& root_element_name,
             xercesc::ErrorHandler& eh)
      {
        xml::auto_initializer initializer;

        xml::sax::std_input_source isrc (is);

        parse (isrc, root_element_name, eh);
      }

      template <typename C>
      void traverser<void, C>::
      parse (std::istream& is,
             std::basic_string<C> const& root_element_name,
             xercesc::SAX2XMLReader& sax)
      {
        xml::auto_initializer initializer;

        xml::sax::std_input_source isrc (is);

        parse (isrc, root_element_name, sax);
      }


      // parse (istream, system_id)
      //


      template <typename C>
      void traverser<void, C>::
      parse (std::istream& is,
             std::basic_string<C> const& system_id,
             std::basic_string<C> const& root_element_name)
      {
        xml::auto_initializer initializer;

        xml::sax::std_input_source isrc (is, system_id);

        parse (isrc, root_element_name);
      }


      template <typename C>
      void traverser<void, C>::
      parse (std::istream& is,
             std::basic_string<C> const& system_id,
             std::basic_string<C> const& root_element_name,
             xercesc::ErrorHandler& eh)
      {
        xml::auto_initializer initializer;

        xml::sax::std_input_source isrc (is, system_id);

        parse (isrc, root_element_name, eh);
      }

      template <typename C>
      void traverser<void, C>::
      parse (std::istream& is,
             std::basic_string<C> const& system_id,
             std::basic_string<C> const& root_element_name,
             xercesc::SAX2XMLReader& sax)
      {
        xml::auto_initializer initializer;

        xml::sax::std_input_source isrc (is, system_id);

        parse (isrc, root_element_name, sax);
      }

      // parse (istream, system_id, public_id)
      //

      template <typename C>
      void traverser<void, C>::
      parse (std::istream& is,
             std::basic_string<C> const& system_id,
             std::basic_string<C> const& public_id,
             std::basic_string<C> const& root_element_name)
      {
        xml::auto_initializer initializer;

        xml::sax::std_input_source isrc (is, system_id, public_id);

        parse (isrc, root_element_name);
      }

      template <typename C>
      void traverser<void, C>::
      parse (std::istream& is,
             std::basic_string<C> const& system_id,
             std::basic_string<C> const& public_id,
             std::basic_string<C> const& root_element_name,
             xercesc::ErrorHandler& eh)
      {
        xml::auto_initializer initializer;

        xml::sax::std_input_source isrc (is, system_id, public_id);

        parse (isrc, root_element_name, eh);
      }

      template <typename C>
      void traverser<void, C>::
      parse (std::istream& is,
             std::basic_string<C> const& system_id,
             std::basic_string<C> const& public_id,
             std::basic_string<C> const& root_element_name,
             xercesc::SAX2XMLReader& sax)
      {
        xml::auto_initializer initializer;

        xml::sax::std_input_source isrc (is, system_id, public_id);

        parse (isrc, root_element_name, sax);
      }


      // parse (InputSource)
      //


      template <typename C>
      void traverser<void, C>::
      parse (xercesc::InputSource const& is,
             std::basic_string<C> const& root_element_name)
      {
        std::auto_ptr<xercesc::SAX2XMLReader> sax (create_sax ());

        parse (is, root_element_name, *sax);
      }


      template <typename C>
      void traverser<void, C>::
      parse (xercesc::InputSource const& is,
             std::basic_string<C> const& root_element_name,
             xercesc::ErrorHandler& eh)
      {
        std::auto_ptr<xercesc::SAX2XMLReader> sax (create_sax ());

        sax->setErrorHandler (&eh);

        parse (is, root_element_name, *sax);
      }


      template <typename C>
      void traverser<void, C>::
      parse (xercesc::InputSource const& is,
             std::basic_string<C> const& root_element_name,
             xercesc::SAX2XMLReader& sax)
      {
        using namespace xercesc;

        document<C> doc (*this, root_element_name);
        sax_event_router<C> router (doc);

        ErrorHandler* old_eh (sax.getErrorHandler ());
        xml::sax::error_handler<C> eh (old_eh);

        try
        {
          sax.setErrorHandler (&eh);

          sax.setContentHandler (&router);
          sax.parse (is);

          sax.setErrorHandler (old_eh);
        }
        catch (...)
        {
          sax.setErrorHandler (old_eh);
          throw;
        }

        if (eh.failed ())
          throw invalid_instance ();

        post ();
      }


      template <typename C>
      std::auto_ptr<xercesc::SAX2XMLReader> traverser<void, C>::
      create_sax ()
      {
        using namespace xercesc;

        std::auto_ptr<SAX2XMLReader> sax (
          XMLReaderFactory::createXMLReader());

        sax->setFeature(XMLUni::fgSAX2CoreValidation, true);
        sax->setFeature(XMLUni::fgXercesDynamic, false);
        sax->setFeature(XMLUni::fgSAX2CoreNameSpaces, true);
        sax->setFeature(XMLUni::fgXercesSchema, true);
        sax->setFeature(XMLUni::fgXercesSchemaFullChecking, true);
        sax->setFeature(XMLUni::fgSAX2CoreNameSpacePrefixes, true);
        sax->setFeature (XMLUni::fgXercesValidationErrorAsFatal, true);

        return sax;
      }
    }
  }
}
