// file      : xsd/cxx/xml/dom/elements.txx
// author    : Boris Kolpackov <boris@codesynthesis.com>
// copyright : Copyright (c) 2005 Code Synthesis Tools CC
// license   : GNU GPL v2 + exceptions; see accompanying LICENSE file

#include <xercesc/util/XMLUniDefs.hpp> // chLatin_L, etc

#include <xsd/cxx/xml/dom/error-handler.hxx>

namespace xsd
{
  namespace cxx
  {
    namespace xml
    {
      namespace dom
      {
      
template <typename C>
xsd::cxx::xml::dom::auto_ptr<xercesc::DOMDocument>
dom (xercesc::DOMInputSource const& is, xercesc::DOMErrorHandler* eh)
{
  using namespace xercesc;

  // Instantiate the DOM parser.
  //
  XMLCh const gLS[] = {chLatin_L, chLatin_S, chNull };

  // Get an implementation of the Load-Store (LS) interface.
  //
  DOMImplementationLS* impl (
    static_cast<DOMImplementationLS*>(
      DOMImplementationRegistry::getDOMImplementation(gLS)));

  // Create a DOMBuilder.
  //
  auto_ptr<DOMBuilder> parser (
    impl->createDOMBuilder(DOMImplementationLS::MODE_SYNCHRONOUS, 0));

  // Discard comment nodes in the document.
  //
  parser->setFeature (XMLUni::fgDOMComments, false);

  // Disable datatype normalization. The XML 1.0 attribute value
  // normalization always occurs though.
  //
  parser->setFeature (XMLUni::fgDOMDatatypeNormalization, true);

  // Do not create EntityReference nodes in the DOM tree. No
  // EntityReference nodes will be created, only the nodes
  // corresponding to their fully expanded substitution text will be
  // created.
  //
  parser->setFeature (XMLUni::fgDOMEntities, false);

  // Perform Namespace processing.
  //
  parser->setFeature (XMLUni::fgDOMNamespaces, true);

  // Perform Validation
  //
  parser->setFeature (XMLUni::fgDOMValidation, true);

  // Do not include ignorable whitespace in the DOM tree.
  //
  parser->setFeature (XMLUni::fgDOMWhitespaceInElementContent, false);

  // Enable the parser's schema support.
  //
  parser->setFeature (XMLUni::fgXercesSchema, true);

  // Enable full schema constraint checking, including checking which
  // may be time-consuming or memory intensive. Currently, particle
  // unique attribution constraint checking and particle derivation
  // restriction checking are controlled by this option.
  //
  parser->setFeature (XMLUni::fgXercesSchemaFullChecking, true);

  // The parser will treat validation error as fatal and will exit.
  //
  parser->setFeature (XMLUni::fgXercesValidationErrorAsFatal, true);

  // We will release DOM ourselves.
  //
  parser->setFeature (XMLUni::fgXercesUserAdoptsDOMDocument, true);

  // Set error handler.
  //
  error_handler<C> handler (eh);
  parser->setErrorHandler (&handler);

  auto_ptr<DOMDocument> doc (parser->parse (is));

  if (handler.failed ())
    doc.reset ();

  return doc;
}

template <typename C>
xsd::cxx::xml::dom::auto_ptr<xercesc::DOMDocument>
dom (std::basic_string<C> const& uri, xercesc::DOMErrorHandler* eh)
{
  using namespace xercesc;

  // Instantiate the DOM parser.
  //
  XMLCh const gLS[] = {chLatin_L, chLatin_S, chNull };

  // Get an implementation of the Load-Store (LS) interface.
  //
  DOMImplementationLS* impl (
    static_cast<DOMImplementationLS*>(
      DOMImplementationRegistry::getDOMImplementation(gLS)));

  // Create a DOMBuilder.
  //
  auto_ptr<DOMBuilder> parser (
    impl->createDOMBuilder(DOMImplementationLS::MODE_SYNCHRONOUS, 0));

  // Discard comment nodes in the document.
  //
  parser->setFeature (XMLUni::fgDOMComments, false);

  // Disable datatype normalization. The XML 1.0 attribute value
  // normalization always occurs though.
  //
  parser->setFeature (XMLUni::fgDOMDatatypeNormalization, true);

  // Do not create EntityReference nodes in the DOM tree. No
  // EntityReference nodes will be created, only the nodes
  // corresponding to their fully expanded substitution text will be
  // created.
  //
  parser->setFeature (XMLUni::fgDOMEntities, false);

  // Perform Namespace processing.
  //
  parser->setFeature (XMLUni::fgDOMNamespaces, true);

  // Perform Validation
  //
  parser->setFeature (XMLUni::fgDOMValidation, true);

  // Do not include ignorable whitespace in the DOM tree.
  //
  parser->setFeature (XMLUni::fgDOMWhitespaceInElementContent, false);

  // Enable the parser's schema support.
  //
  parser->setFeature (XMLUni::fgXercesSchema, true);

  // Enable full schema constraint checking, including checking which
  // may be time-consuming or memory intensive. Currently, particle
  // unique attribution constraint checking and particle derivation
  // restriction checking are controlled by this option.
  //
  parser->setFeature (XMLUni::fgXercesSchemaFullChecking, true);

  // The parser will treat validation error as fatal and will exit.
  //
  parser->setFeature (XMLUni::fgXercesValidationErrorAsFatal, true);

  // We will release DOM ourselves.
  //
  parser->setFeature (XMLUni::fgXercesUserAdoptsDOMDocument, true);

  // Set error handler.
  //
  error_handler<C> handler (eh);
  parser->setErrorHandler (&handler);

  auto_ptr<DOMDocument> doc (parser->parseURI (string (uri).c_str ()));

  if (handler.failed ())
    doc.reset ();

  return doc;
}
      }
    }
  }
}
